
/*
 * This is the one and only source of version information.
 *
 * Do not include this file directly, it is available via
 * auto-generated version.c, which also includes the git sha.
 */

#define MP_VERSION_MAJOR 3
#define MP_VERSION_MINOR 5
#define MP_VERSION_PATCH 0

